
/* Router solicitation transmission routines. */


/*************************************************************/
/*
Please note that either this license MUST be displayed
or a particular instance of this code may be converted to
the GNU General Public License v2 or greater.

Please see README.txt for details.
*/

/*
Copyright (c) 2005, Monash University
All rights reserved.
Contributors: Greg Daley

Redistribution and use in source and binary forms, with or
without modification, are permitted provided that the following
conditions are met:

* Redistributions of source code must retain the above copyright
  notice, this list of conditions and the following disclaimer.

* Redistributions in binary form must reproduce the above
  copyright notice, this list of conditions and the following
  disclaimer in the documentation and/or other materials
  provided with the distribution.

* Neither the name of Monash University nor the names of its
  contributors may be used to endorse or promote products derived
  from this software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
"AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY
WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
POSSIBILITY OF SUCH DAMAGE.
 
*/
/*************************************************************/

#include <stdio.h>
#include <netinet/ip6.h>
#include <netinet/icmp6.h> 
#include <errno.h> 


#ifndef ND_OPT_TENTATIVE_SLLA
#define ND_OPT_TENTATIVE_SLLA     0x11
#endif  /* ND_OPT_TENTATIVE_SLLA */

#ifndef ND_OPT_LANDMARK
#define ND_OPT_LANDMARK 0x13

struct nd_opt_landmark{
	uint8_t  nd_opt_landmark_type;
	uint8_t  nd_opt_landmark_len;
	uint8_t  nd_opt_landmark_plen;
	uint8_t  nd_opt_landmark_flags;
	uint32_t nd_opt_landmark_reserved;
	struct in6_addr nd_opt_landmark_prefix;
};

#endif /* ND_OPT_LANDMARK*/


#ifndef MSG_SIZE
#define MSG_SIZE 1500
#endif				/* MSG_SIZE */

uint16_t icmpv6_cksum(unsigned char *icmpmsg,
		      struct in6_addr *src, struct in6_addr *dest,
		      int icmplen);

void
send_rs(int sock, int ifindex, uint8_t macaddr[], int tentative,
	struct in6_addr *src, struct in6_addr *dest, uint8_t *prefix)
{
	struct in6_addr all_routers;
	struct in6_addr unspec;
	struct ip6_hdr *sixh;
	struct nd_router_solicit *rsol;
	unsigned char buff[MSG_SIZE];
	struct sockaddr_in6 sa6;
	int len = 0;

	memset((void *) buff, 0, sizeof(buff));
	memset((void *) &unspec, 0, sizeof(struct in6_addr));
	memset((void *) &all_routers, 0, sizeof(struct in6_addr));

	all_routers.s6_addr[0] = 0xff;
	all_routers.s6_addr[1] = all_routers.s6_addr[15] = 0x02;


	if (dest == NULL) {
		dest = (struct in6_addr *) &all_routers;
	}
	if (src == NULL) {
		src = (struct in6_addr *) &unspec;
	}

	sixh = (struct ip6_hdr *) buff;

	sixh->ip6_vfc  = 6 << 4;
	sixh->ip6_hlim = 0xff;
	sixh->ip6_nxt  = IPPROTO_ICMPV6;

	memcpy(&sixh->ip6_dst, dest, sizeof(struct in6_addr));
	memcpy(&sixh->ip6_src, src,  sizeof(struct in6_addr));

	len = sizeof(struct ip6_hdr);

	rsol = (struct nd_router_solicit *) (buff + len);

	rsol->nd_rs_type = ND_ROUTER_SOLICIT;
	rsol->nd_rs_code = 0;
	rsol->nd_rs_cksum = 0;

	/*printf("creating RS\n");*/
	len += sizeof(struct nd_router_solicit);

	/*
	 * add (Tentative?) Source Link-layer Address option
	 */

	if (memcmp(&unspec, &sixh->ip6_src, sizeof(struct in6_addr))) {
		uint8_t *ucp;

		ucp = (uint8_t *) (buff + len);

		*ucp++ = (tentative ? ND_OPT_TENTATIVE_SLLA :
			  ND_OPT_SOURCE_LINKADDR);
		*ucp++ = (uint8_t) 1;

		len += 2 * sizeof(uint8_t);

		memcpy(buff + len, macaddr, 6);
		len += 6;
	}

	/* assuming not 000/3! */
	if(memcmp(prefix, &unspec, 8)){
		buff[len] = ND_OPT_LANDMARK;
		buff[len+1]= 3 ;
		buff[len+2]=64; /* not the best way to code it ...*/
		printf("Land: %2.2x%2.2x:%2.2x%2.2x:%2.2x%2.2x:%2.2x%2.2x\n",
			prefix[0], prefix[1], prefix[2], prefix[3],
			prefix[4], prefix[5], prefix[6], prefix[7]);
		memcpy(&buff[len+8], prefix, 8);
		
		printf("Land: %2.2x%2.2x:%2.2x%2.2x:%2.2x%2.2x:%2.2x%2.2x\n",
			buff[len+8], buff[len+9], buff[len+10], buff[len+11],
			buff[len+12], buff[len+13], buff[len+14], buff[len+15]);
		len+= sizeof(struct nd_opt_landmark);
	}

	sixh->ip6_plen = htons((len - sizeof(struct ip6_hdr)));

	rsol->nd_rs_cksum = icmpv6_cksum((unsigned char *) rsol,
					 &sixh->ip6_src, &sixh->ip6_dst,
					 ntohs(sixh->ip6_plen));

	printf("sending RS len: %d ck:(%hx)\n", len, rsol->nd_rs_cksum);

	sa6.sin6_family = AF_INET6;
	sa6.sin6_port = 0;
	sa6.sin6_scope_id = ifindex;
	memcpy(&sa6.sin6_addr, dest, sizeof(struct in6_addr));

	if (sendto(sock, buff, len, 0, (struct sockaddr *) &sa6,
		   sizeof(struct sockaddr_in6)) < 0) {
		fprintf(stderr, "sendto: %s\n", strerror(errno));
	}
}

uint16_t icmpv6_cksum(unsigned char *icmpmsg,
		      struct in6_addr *src, struct in6_addr *dest,
		      int icmplen)
{
	uint64_t workspace = 0;
	struct icmp6_hdr tmphdr;
	int i;
	memcpy(&tmphdr, (struct icmp6_hdr *) icmpmsg,
	       sizeof(struct icmp6_hdr));

	tmphdr.icmp6_cksum = 0;

	// calculate pseudo ipv6  hdr checksum

	for (i = 0; i < 4; i++) {
		workspace += src->s6_addr32[i];
		workspace += dest->s6_addr32[i];
	}
	workspace += htonl(icmplen);
	workspace += htonl(IPPROTO_ICMPV6);

	// calculate pseudo icmpv6 hdr checksum
	workspace += ((uint32_t *) & tmphdr)[0];
	workspace += tmphdr.icmp6_data32[0];

	// calculate across rest of icmpv6 message

	//Assuming end on 32 bit boundary
	for (i = sizeof(struct icmp6_hdr); i < icmplen; i += 4) {
		workspace += *((uint32_t *) & (icmpmsg[i]));
	}
	// merge down 64 bits to 32 bits. 
	workspace =
	    (workspace & 0xffffffff) + ((workspace >> 32) & 0xfffffff);
	workspace =
	    (workspace & 0xffffffff) + ((workspace >> 32) & 0xfffffff);

	// merge down 32 bits to 16 bits
	workspace = (workspace & 0xffff) + ((workspace >> 16) & 0xffff);
	workspace = (workspace & 0xffff) + ((workspace >> 16) & 0xffff);
	// we have preserved all the bits.
	return ~((uint16_t) workspace);
}
